#include <err.h>
#include <fcntl.h>
#include <mqueue.h>
#include <openssl/rand.h>
#include <sched.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/mman.h>
#include <sys/prctl.h>
#include <sys/stat.h>
#include <unistd.h>

#include "interface.h"

static volatile char *shared_mem;
static int64_t balance = 0;

void send_flag() {
  FILE *file = fopen("/home/server/flag.txt", "r");
  if (!file) {
    err(EXIT_FAILURE, "Failed to read flag.txt");
  }

  fread(SHM_ADDR, 1, 128, file);

  struct server_client_message msg = {.type = S2C_FLAG,
                                      .message.flag = SHM_ADDR};
  send(msg, "Failed to send flag");

  if (fclose(file)) {
    err(EXIT_FAILURE, "Failed to close flag.txt");
  }
}

static bool flipped = false;
static bool heads = false;

void dispatch(struct client_server_message *msg) {
  struct server_client_message response = {0};

  unsigned char random;

  uint64_t message;

  switch (msg->type) {
  case C2S_BALANCE:
    response.type = S2C_BALANCE;
    response.message.balance = balance;
    send(response, "Failed to send balance");
    break;

  case C2S_BUY_FLAG:
    if (balance >= FLAG_PRICE) {
      sched_yield();
      balance -= FLAG_PRICE;
      send_flag();
    } else {
      errx(
          EXIT_FAILURE,
          "Tried to buy the flag, but insufficient balance. Exiting server...");
    }
    break;

  case C2S_REQUEST_FLIP:
    if (!flipped) {
      if (RAND_priv_bytes(&random, 1) != 1) {
        err(EXIT_FAILURE, "Failed to get random byte");
      }
    }

    flipped = true;
    heads = random % 2;
    message = *((uint64_t *)&(shared_mem[heads * 4096]));

    response.type = S2C_FLIPPED;
    send(response, "Failed to send request flip response");
    break;

  case C2S_FLIP_PREDICTION:
    if (!flipped) {
      errx(EXIT_FAILURE,
           "Tried to predict without flipping. Exiting server...");
    }
    flipped = false;

    if (msg->message.prediction.bet < 0 ||
        msg->message.prediction.bet > balance || balance < 0) {
      errx(EXIT_FAILURE, "Tried to bet more or less than allowed");
    }

    enum coin_flip_result p = msg->message.prediction.prediction;
    if (p != HEADS && p != TAILS) {
      errx(EXIT_FAILURE, "Predicted an unknown coin state");
    }

    struct server_client_message resp = {.type = S2C_PREDICTION_RESULT};
    if ((p == HEADS && heads) || (p == TAILS && !heads)) {
      balance += msg->message.prediction.bet;
      resp.message.prediction_result.correct = true;
    } else {
      balance -= msg->message.prediction.bet;
      resp.message.prediction_result.correct = false;
    }
    warnx("coin flip was %s", (char *)&message);
    send(resp, "Failed to send coin flip prediction result");
    break;

  default:
    errx(EXIT_FAILURE, "Unknown message type: %d", msg->type);
    break;
  }
}

int main() {
  if (prctl(PR_SET_PDEATHSIG, SIGTERM) == -1) {
    err(EXIT_FAILURE, "Failed to set parent death signal");
  }
  if (getppid() == 1) {
    exit(1);
  }

  // Set the starting balance
  balance = 10;

  setup_message_queues_server();
  shared_mem = setup_shared_memory_server();

  while (1) {
    struct client_server_message message;
    recv(message, "Failed to receive message from client");
    dispatch(&message);
  }
}
